# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os

import pytest
from pymssql import Connection

from airflow.providers.microsoft.mssql.hooks.mssql import MsSqlHook

AIRFLOW_CONN_MSSQL_DEFAULT = "mssql://sa:airflow123@mssql:1433/"


@pytest.mark.integration("mssql")
class TestMsSqlHook:
    def setup_method(self):
        os.environ["AIRFLOW_CONN_MSSQL_DEFAULT"] = AIRFLOW_CONN_MSSQL_DEFAULT
        self.hook = MsSqlHook()

    def test_get_conn(self):
        assert isinstance(self.hook.get_conn(), Connection)

    def test_autocommit(self):
        conn = self.hook.get_conn()
        self.hook.set_autocommit(conn=conn, autocommit=True)
        assert self.hook.get_autocommit(conn=conn)
        self.hook.set_autocommit(conn=conn, autocommit=False)
        assert not self.hook.get_autocommit(conn=conn)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import re
from os.path import dirname, join


def remove_license_header(content: str) -> str:
    """Remove license header from the given content."""
    # Define the pattern to match both block and single-line comments
    pattern = r"(/\*.*?\*/)|(--.*?(\r?\n|\r))|(#.*?(\r?\n|\r))"

    # Check if there is a license header at the beginning of the file
    if re.match(pattern, content, flags=re.DOTALL):
        # Use re.DOTALL to allow .* to match newline characters in block comments
        return re.sub(pattern, "", content, flags=re.DOTALL).strip()
    return content.strip()


def load_file(*args: str, mode="r", encoding="utf-8"):
    with open(join(dirname(__file__), *args), mode=mode, encoding=encoding) as file:
        if mode == "r":
            return remove_license_header(file.read())
        return file.read()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.providers.microsoft.mssql.dialects.mssql import MsSqlDialect


class TestMsSqlDialect:
    @pytest.mark.parametrize(
        "create_db_api_hook",
        [
            (
                [
                    {"name": "index", "identity": True},
                    {"name": "name"},
                    {"name": "firstname"},
                    {"name": "age"},
                ],  # columns
                [("index",)],  # primary_keys
                {"index", "user"},  # reserved_words
                False,  # escape_column_names
            ),
        ],
        indirect=True,
    )
    def test_placeholder(self, create_db_api_hook):
        assert MsSqlDialect(create_db_api_hook).placeholder == "?"

    @pytest.mark.parametrize(
        "create_db_api_hook",
        [
            (
                [
                    {"name": "index", "identity": True},
                    {"name": "name"},
                    {"name": "firstname"},
                    {"name": "age"},
                ],  # columns
                [("index",)],  # primary_keys
                {"index", "user"},  # reserved_words
                False,  # escape_column_names
            ),
        ],
        indirect=True,
    )
    def test_get_column_names(self, create_db_api_hook):
        assert MsSqlDialect(create_db_api_hook).get_column_names("hollywood.actors") == [
            "index",
            "name",
            "firstname",
            "age",
        ]

    @pytest.mark.parametrize(
        "create_db_api_hook",
        [
            (
                [
                    {"name": "index", "identity": True},
                    {"name": "name"},
                    {"name": "firstname"},
                    {"name": "age"},
                ],  # columns
                [("index",)],  # primary_keys
                {"index", "user"},  # reserved_words
                False,  # escape_column_names
            ),
        ],
        indirect=True,
    )
    def test_get_target_fields(self, create_db_api_hook):
        assert MsSqlDialect(create_db_api_hook).get_target_fields("hollywood.actors") == [
            "name",
            "firstname",
            "age",
        ]

    @pytest.mark.parametrize(
        "create_db_api_hook",
        [
            (
                [
                    {"name": "index", "identity": True},
                    {"name": "name"},
                    {"name": "firstname"},
                    {"name": "age"},
                ],  # columns
                [("index",)],  # primary_keys
                {"index", "user"},  # reserved_words
                False,  # escape_column_names
            ),
        ],
        indirect=True,
    )
    def test_get_primary_keys(self, create_db_api_hook):
        assert MsSqlDialect(create_db_api_hook).get_primary_keys("hollywood.actors") == ["index"]

    @pytest.mark.parametrize(
        "create_db_api_hook",
        [
            (
                [
                    {"name": "index", "identity": True},
                    {"name": "name"},
                    {"name": "firstname"},
                    {"name": "age"},
                ],  # columns
                [("index",)],  # primary_keys
                {"index", "user"},  # reserved_words
                False,  # escape_column_names
            ),
        ],
        indirect=True,
    )
    def test_generate_replace_sql(self, create_db_api_hook):
        values = [
            {"index": 1, "name": "Stallone", "firstname": "Sylvester", "age": "78"},
            {"index": 2, "name": "Statham", "firstname": "Jason", "age": "57"},
            {"index": 3, "name": "Li", "firstname": "Jet", "age": "61"},
            {"index": 4, "name": "Lundgren", "firstname": "Dolph", "age": "66"},
            {"index": 5, "name": "Norris", "firstname": "Chuck", "age": "84"},
        ]
        target_fields = ["index", "name", "firstname", "age"]
        sql = MsSqlDialect(create_db_api_hook).generate_replace_sql("hollywood.actors", values, target_fields)
        assert (
            sql
            == """
            MERGE INTO hollywood.actors WITH (ROWLOCK) AS target
            USING (SELECT ? AS [index], ? AS name, ? AS firstname, ? AS age) AS source
            ON target.[index] = source.[index]
            WHEN MATCHED THEN
                UPDATE SET target.name = source.name, target.firstname = source.firstname, target.age = source.age
            WHEN NOT MATCHED THEN
                INSERT ([index], name, firstname, age) VALUES (source.[index], source.name, source.firstname, source.age);
        """.strip()
        )

    @pytest.mark.parametrize(
        "create_db_api_hook",
        [
            (
                [
                    {"name": "index", "identity": True},
                    {"name": "name", "identity": True},
                    {"name": "firstname", "identity": True},
                    {"name": "age", "identity": True},
                ],  # columns
                [("index",), ("name",), ("firstname",), ("age",)],  # primary_keys
                {"index", "user"},  # reserved_words
                False,  # escape_column_names
            ),
        ],
        indirect=True,
    )
    def test_generate_replace_sql_when_all_columns_are_part_of_primary_key(self, create_db_api_hook):
        values = [
            {"index": 1, "name": "Stallone", "firstname": "Sylvester", "age": "78"},
            {"index": 2, "name": "Statham", "firstname": "Jason", "age": "57"},
            {"index": 3, "name": "Li", "firstname": "Jet", "age": "61"},
            {"index": 4, "name": "Lundgren", "firstname": "Dolph", "age": "66"},
            {"index": 5, "name": "Norris", "firstname": "Chuck", "age": "84"},
        ]
        target_fields = ["index", "name", "firstname", "age"]
        sql = MsSqlDialect(create_db_api_hook).generate_replace_sql("hollywood.actors", values, target_fields)
        assert (
            sql
            == """
            MERGE INTO hollywood.actors WITH (ROWLOCK) AS target
            USING (SELECT ? AS [index], ? AS name, ? AS firstname, ? AS age) AS source
            ON target.[index] = source.[index] AND target.name = source.name AND target.firstname = source.firstname AND target.age = source.age
            WHEN NOT MATCHED THEN
                INSERT ([index], name, firstname, age) VALUES (source.[index], source.name, source.firstname, source.age);
        """.strip()
        )

    @pytest.mark.parametrize(
        "create_db_api_hook",
        [
            (
                [
                    {"name": "index", "identity": True},
                    {"name": "name"},
                    {"name": "firstname"},
                    {"name": "age"},
                ],  # columns
                [("index",)],  # primary_keys
                {"index", "user"},  # reserved_words
                True,  # escape_column_names
            ),
        ],
        indirect=True,
    )
    def test_generate_replace_sql_when_escape_column_names_is_enabled(self, create_db_api_hook):
        values = [
            {"index": 1, "name": "Stallone", "firstname": "Sylvester", "age": "78"},
            {"index": 2, "name": "Statham", "firstname": "Jason", "age": "57"},
            {"index": 3, "name": "Li", "firstname": "Jet", "age": "61"},
            {"index": 4, "name": "Lundgren", "firstname": "Dolph", "age": "66"},
            {"index": 5, "name": "Norris", "firstname": "Chuck", "age": "84"},
        ]
        target_fields = ["index", "name", "firstname", "age"]
        sql = MsSqlDialect(create_db_api_hook).generate_replace_sql("hollywood.actors", values, target_fields)
        assert (
            sql
            == """
            MERGE INTO hollywood.actors WITH (ROWLOCK) AS target
            USING (SELECT ? AS [index], ? AS [name], ? AS [firstname], ? AS [age]) AS source
            ON target.[index] = source.[index]
            WHEN MATCHED THEN
                UPDATE SET target.[name] = source.[name], target.[firstname] = source.[firstname], target.[age] = source.[age]
            WHEN NOT MATCHED THEN
                INSERT ([index], [name], [firstname], [age]) VALUES (source.[index], source.[name], source.[firstname], source.[age]);
        """.strip()
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example use of MsSql related operators.
"""

from __future__ import annotations

# [START mssql_operator_howto_guide]
import os
from datetime import datetime

import pytest

from airflow import DAG

try:
    from airflow.providers.common.sql.operators.sql import SQLExecuteQueryOperator
    from airflow.providers.microsoft.mssql.hooks.mssql import MsSqlHook
except ImportError:
    pytest.skip("MSSQL provider not available", allow_module_level=True)

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID")
DAG_ID = "example_mssql"


with DAG(
    DAG_ID,
    schedule="@daily",
    start_date=datetime(2021, 10, 1),
    tags=["example"],
    catchup=False,
) as dag:
    # [START howto_operator_mssql]

    # Example of creating a task to create a table in MsSql

    create_table_mssql_task = SQLExecuteQueryOperator(
        task_id="create_country_table",
        conn_id="airflow_mssql",
        sql=r"""
        CREATE TABLE Country (
            country_id INT NOT NULL IDENTITY(1,1) PRIMARY KEY,
            name TEXT,
            continent TEXT
        );
        """,
        dag=dag,
    )

    # [END howto_operator_mssql]

    # [START mssql_hook_howto_guide_insert_mssql_hook]

    @dag.task(task_id="insert_mssql_task")
    def insert_mssql_hook():
        mssql_hook = MsSqlHook(mssql_conn_id="airflow_mssql", schema="airflow")

        rows = [
            ("India", "Asia"),
            ("Germany", "Europe"),
            ("Argentina", "South America"),
            ("Ghana", "Africa"),
            ("Japan", "Asia"),
            ("Namibia", "Africa"),
        ]
        target_fields = ["name", "continent"]
        mssql_hook.insert_rows(table="Country", rows=rows, target_fields=target_fields)

    # [END mssql_hook_howto_guide_insert_mssql_hook]

    # [START mssql_operator_howto_guide_create_table_mssql_from_external_file]
    # Example of creating a task that calls an sql command from an external file.
    create_table_mssql_from_external_file = SQLExecuteQueryOperator(
        task_id="create_table_from_external_file",
        conn_id="airflow_mssql",
        sql="create_table.sql",
        dag=dag,
    )
    # [END mssql_operator_howto_guide_create_table_mssql_from_external_file]

    # [START mssql_operator_howto_guide_populate_user_table]
    populate_user_table = SQLExecuteQueryOperator(
        task_id="populate_user_table",
        conn_id="airflow_mssql",
        sql=r"""
                INSERT INTO Users (username, description)
                VALUES ( 'Danny', 'Musician');
                INSERT INTO Users (username, description)
                VALUES ( 'Simone', 'Chef');
                INSERT INTO Users (username, description)
                VALUES ( 'Lily', 'Florist');
                INSERT INTO Users (username, description)
                VALUES ( 'Tim', 'Pet shop owner');
                """,
    )
    # [END mssql_operator_howto_guide_populate_user_table]

    # [START mssql_operator_howto_guide_get_all_countries]
    get_all_countries = SQLExecuteQueryOperator(
        task_id="get_all_countries",
        conn_id="airflow_mssql",
        sql=r"""SELECT * FROM Country;""",
    )
    # [END mssql_operator_howto_guide_get_all_countries]

    # [START mssql_operator_howto_guide_get_all_description]
    get_all_description = SQLExecuteQueryOperator(
        task_id="get_all_description",
        conn_id="airflow_mssql",
        sql=r"""SELECT description FROM Users;""",
    )
    # [END mssql_operator_howto_guide_get_all_description]

    # [START mssql_operator_howto_guide_params_passing_get_query]
    get_countries_from_continent = SQLExecuteQueryOperator(
        task_id="get_countries_from_continent",
        conn_id="airflow_mssql",
        sql=r"""SELECT * FROM Country where {{ params.column }}='{{ params.value }}';""",
        params={"column": "CONVERT(VARCHAR, continent)", "value": "Asia"},
    )
    # [END mssql_operator_howto_guide_params_passing_get_query]
    (
        create_table_mssql_task
        >> insert_mssql_hook()
        >> create_table_mssql_from_external_file
        >> populate_user_table
        >> get_all_countries
        >> get_all_description
        >> get_countries_from_continent
    )
    # [END mssql_operator_howto_guide]

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()
from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)
